@@ -1,13 +1,15 @@
-from framework.network import MultiLayerNet
+from jjy.framework.network import MultiLayerNet
 import random
-from dataset.mnist import load_mnist
-import numpy as np
+from jjy.dataset.mnist import load_mnist
+from jjy.framework.cuda import xp as np
 import matplotlib.pyplot as plt
 
 (x_train, t_train), (x_test, t_test) = load_mnist(flatten=False)
 
 net = MultiLayerNet()
-net.load_model("train_weight_2021-04-10 213538.npz")
+net.load_model("train_weight_2021-04-18 151635.npz")
+
+
 
 for i in range(5):
     img_idx = random.randrange(0, 10000)
