@@ -1,14 +1,15 @@
 # %%
 import jsonpickle as jsp
 from inspect import currentframe
-
+import numpy as np2
 import jjy.framework.initializer as Initializer
 import jjy.framework.layer as Layer
 import jjy.framework.optimizer as Optimizer
 from jjy.framework.functions import *
 from tqdm import tqdm
-
+import os
 import json
+import math
 
 
 def debug_print(arg):
@@ -26,7 +27,8 @@ sys.path.append("../../")
 
 # %%
 import re
-import numpy as np
+from jjy.framework.cuda import xp as np
+from jjy.framework.cuda import GPU_ENABLE, GPU_MODE
 from collections import OrderedDict
 
 
@@ -37,6 +39,8 @@ class MultiLayerNet:
 
     def __init__(self, weight_decay_lambda=0, is_use_dropout=False, dropout_ratio=0.5):
 
+        self.is_gpu = False
+
         self.weight_decay_lambda = weight_decay_lambda
         self.is_use_dropout = is_use_dropout
         self.dropout_ratio = dropout_ratio
@@ -76,7 +80,7 @@ class MultiLayerNet:
             if input_size is None and len(self.hiddenSizeList) > 0:
                 #                 input_size = self.prevDenseLayer.hidden_size
                 #                 print(self.hiddenSizeList)
-                input_size = int(np.prod(self.hiddenSizeList[-1]))
+                input_size = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
             else:
                 self.hiddenSizeList.append(input_size)
 
@@ -94,8 +98,6 @@ class MultiLayerNet:
             elif isinstance(initializer, Initializer.Xavier):
                 weight_init_std = np.sqrt(1.0 / pre_node_nums)
 
-
-
             self.hiddenSizeList.append(hidden_size)
             #             print(input_size)
             #             print("input", self.hiddenSizeList)
@@ -133,9 +135,9 @@ class MultiLayerNet:
             initializer = layer.initializer
 
             if self.prevConvLayer is None:
-                pre_node_nums = np.prod(layer.filter_size)
+                pre_node_nums = np.prod(np.asarray(layer.filter_size))
             else:
-                pre_node_nums = self.prevConvLayer.filter_num * np.prod(layer.filter_size)
+                pre_node_nums = self.prevConvLayer.filter_num * np.prod(np.asarray(layer.filter_size))
 
             if isinstance(initializer, Initializer.Std):
                 weight_init_std = initializer.std
@@ -150,8 +152,8 @@ class MultiLayerNet:
             # print(layer.filter_size, layer.pad, layer.stride, input_size, "mu")
 
             conv_output_size = (
-                layer.filter_num, (input_size[1] - layer.filter_size[0] + 2 * layer.pad) / layer.stride + 1,
-                (input_size[2] - layer.filter_size[1] + 2 * layer.pad) / layer.stride + 1)
+                layer.filter_num, int((input_size[1] - layer.filter_size[0] + 2 * layer.pad) / layer.stride + 1),
+                int((input_size[2] - layer.filter_size[1] + 2 * layer.pad) / layer.stride + 1))
             #
             self.hiddenSizeList.append(conv_output_size)
 
@@ -220,11 +222,26 @@ class MultiLayerNet:
 
 
         elif isinstance(layer, Layer.BatchNormalization):
+            # print(self.hiddenSizeList)
+            # print(np.prod(np.asarray([self.hiddenSizeList[-1]])))
+            # if type(self.hiddenSizeList[-1]) == int:
+            #     n = self.hiddenSizeList[-1]
+            # else:
+            #     n = self.prevConvLayer.filter_num
 
-            self.layers[f"BatchNormal{layer_len}"] = Layer.BatchNormalization(
-                gamma=np.ones(self.hiddenSizeList[-1]),
-                beta=np.zeros(self.hiddenSizeList[-1])
-            )
+            n = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
+            # print(n)
+            layer.__init__(gamma=np.ones(n),
+                beta=np.zeros(n),
+                running_mean=layer.running_mean,
+                running_var=layer.running_var)
+            self.layers[f"BatchNormal{layer_len}"] = layer
+            # self.layers[f"BatchNormal{layer_len}"] = Layer.BatchNormalization(
+            #     gamma=np.ones(n),
+            #     beta=np.zeros(n),
+            #     running_mean=layer.running_mean,
+            #     running_var=layer.running_var
+            # )
         # print(layer, self.hiddenSizeList)
         if is_direct:
             self.added_layer_list.append(layer)
@@ -254,13 +271,25 @@ class MultiLayerNet:
 
         return self.lastLayer.forward(y, t) + weight_decay
 
-    def accuracy(self, x, t):
+    def accuracy(self, x, t, batch_size=100):
 
-        y = self.predict(x, train_flg=False)
-        y = np.argmax(y, axis=1)
-        if t.ndim != 1: t = np.argmax(t, axis=1)
+        correct_cnt = 0
+        data_num = int(x.shape[0])
 
-        accuracy = np.sum(y == t) / float(x.shape[0])
+        for i in range(math.ceil(data_num / batch_size)):
+            # print("batch", i)
+            xx = x[batch_size * i: min(batch_size * (i+1), data_num)]
+            tt = t[batch_size * i: min(batch_size * (i+1), data_num)]
+
+            yy = self.predict(xx, train_flg=False)
+            yy = np.argmax(yy, axis=1)
+            if tt.ndim != 1: tt = np.argmax(tt, axis=1)
+
+            correct_cnt += np.sum(yy == tt)
+
+
+
+        accuracy = correct_cnt / float(x.shape[0])
         return accuracy
 
     def gradient(self, x, t):
@@ -293,6 +322,14 @@ class MultiLayerNet:
 
     def train(self, x_train, t_train, x_test, t_test, **kwargs):  # t_data 는 원 핫 인코딩
 
+        if GPU_ENABLE:
+            if GPU_MODE:
+                print("GPU로 동작합니다. CPU로 동작하게 하려면 'JJY_GPU_ENABLE' 환경변수의 값을 'false'로 설정해주세요")
+            else:
+                print("CPU로 동작합니다. GPU로 동작하게 하려면 cupy 라이브러리를 설치해주세요.")
+        else:
+            print("CPU로 동작합니다. GPU로 동작하게 하려면 'JJY_GPU_ENABLE' 환경변수의 값을 'true'로 설정해주세요")
+
         optimizer = Optimizer.SGD(lr=0.01)
         if "optimizer" in kwargs:
             if isinstance(kwargs["optimizer"], Optimizer.OptimizerType):
@@ -300,47 +337,42 @@ class MultiLayerNet:
             else:
                 raise "err"
 
-        iters_num = 10000
-        if "iters_num" in kwargs:
-            if 0 < kwargs["iters_num"]:
-                iters_num = kwargs["iters_num"]
-            else:
-                raise "err"
+        iters_num = kwargs.get("iters_num", None)
+        if iters_num is None:
+            return ValueError("iters_num must be set")
+        elif not (iters_num > 0):
+            raise ValueError("iters_num must be > 0 ")
 
-        batch_size = 100
-        if "batch_size" in kwargs:
-            if 0 < kwargs["batch_size"]:
-                batch_size = kwargs["batch_size"]
-            else:
-                raise "err"
+        batch_size = kwargs.get("batch_size", None)
+        if batch_size is None:
+            return ValueError("batch_size must be set")
+        elif not (batch_size > 0):
+            raise ValueError("batch_size must be > 0")
 
-        print_epoch = 1
-        if "print_epoch" in kwargs:
-            if kwargs["print_epoch"] is None:
-                print_epoch = 9999999999999999999
+        print_epoch = kwargs.get("print_epoch", 1)
+        if print_epoch is None:
+            print_epoch = 9999999999999999999
+        elif not (print_epoch > 0):
+            raise ValueError("print_epoch must be > 0")
 
-            elif kwargs["print_epoch"] > 0:
-                print_epoch = kwargs["print_epoch"]
-            else:
-                raise "err"
+        save_model_each_epoch = kwargs.get("save_model_each_epoch", None)
+        if save_model_each_epoch is None:
+            save_model_each_epoch = 9999999999999999999
+        elif not (save_model_each_epoch > 0):
+            raise ValueError("save_model_each_epoch must be > 0")
 
-        evaluate_limit = None
-        if "evaluate_limit" in kwargs:
-            if 0 < kwargs["evaluate_limit"]:
-                evaluate_limit = kwargs["evaluate_limit"]
-            else:
-                raise "err"
+        evaluate_limit = kwargs.get("evaluate_limit", None)
+        save_model_path = kwargs.get("save_model_path", "")
 
-        is_use_progress_bar = False
-        if "is_use_progress_bar" in kwargs:
-            is_use_progress_bar = kwargs["is_use_progress_bar"]
+        is_use_progress_bar = kwargs.get("is_use_progress_bar", False)
 
         e_x_train, e_x_test, e_t_train, e_t_test = x_train, x_test, t_train, t_test
 
         if evaluate_limit is not None:
-            e_x_train, e_x_test, e_t_train, e_t_test = x_train[:evaluate_limit], x_test[:evaluate_limit], t_train[
-                                                                                                          :evaluate_limit], t_test[
-                                                                                                                            :evaluate_limit]
+            e_x_train, e_x_test, e_t_train, e_t_test = np.asarray(x_train[:evaluate_limit]), np.asarray(
+                x_test[:evaluate_limit]), np.asarray(t_train[
+                                                     :evaluate_limit]), np.asarray(t_test[
+                                                                                   :evaluate_limit])
 
         output_type = "regression"
         if type(self.lastLayer) == type(Layer.SoftmaxWithLoss()):
@@ -374,9 +406,9 @@ class MultiLayerNet:
 
             # print(i)
 
-            batch_mask = np.random.choice(train_size, batch_size)
-            x_batch = x_train[batch_mask]
-            t_batch = t_train[batch_mask]
+            batch_mask = np2.random.choice(train_size, batch_size)
+            x_batch = np.asarray(x_train[batch_mask])
+            t_batch = np.asarray(t_train[batch_mask])
 
             grad = self.gradient(x_batch, t_batch)
 
@@ -389,27 +421,32 @@ class MultiLayerNet:
 
             if int(cnt * iter_per_epoch) == i:
 
+                if save_model_each_epoch is not None and cnt % save_model_each_epoch == 0:
+                    self.save_model(
+                        save_model_path + f"/train_weight_{str(datetime.datetime.now())[:-7].replace(':', '')}.npz")
+
                 if cnt % print_epoch == 0:
 
-                    train_acc = self.accuracy(e_x_train, e_t_train)
-                    test_acc = self.accuracy(e_x_test, e_t_test)
+                    train_acc = self.accuracy(e_x_train, e_t_train, batch_size=batch_size)
+                    test_acc = self.accuracy(e_x_test, e_t_test, batch_size=batch_size)
                     train_acc_list.append(train_acc)
                     test_acc_list.append(test_acc)
                     if output_type == "regression":
                         print(f"epoch {cnt}:", train_loss_list[-1])
                     else:
                         # log.info(f"epoch {cnt}:")
-                        print(f"epoch {cnt}:", format(train_acc, ".4f"), " | ", format(test_acc, ".4f"), flush=True)
+                        print(f"epoch {cnt}:", format(float(train_acc), ".4f"), " | ", format(float(test_acc), ".4f"),
+                              flush=True)
 
                 cnt += 1
 
         print("===================")
-        train_acc = self.accuracy(e_x_train, e_t_train)
-        test_acc = self.accuracy(e_x_test, e_t_test)
+        train_acc = self.accuracy(e_x_train, e_t_train, batch_size=batch_size)
+        test_acc = self.accuracy(e_x_test, e_t_test, batch_size=batch_size)
         if output_type == "regression":
             print(f"epoch {cnt - 1}:", train_loss_list[-1])
         else:
-            print(f"epoch {cnt - 1}:", format(train_acc, ".4f"), " | ", format(test_acc, ".4f"))
+            print(f"epoch {cnt - 1}:", format(float(train_acc), ".4f"), " | ", format(float(test_acc), ".4f"))
 
         return {
             "train_loss_list": train_loss_list,
@@ -426,41 +463,71 @@ class MultiLayerNet:
             "is_use_dropout": self.is_use_dropout,
             "dropout_ratio": self.dropout_ratio
         }
+
+        other_params = {}
+
         layer_info = []
+
         for layer in self.added_layer_list:
-            s = jsp.encode(layer)
+            t = layer
+            if isinstance(t, Layer.SoftmaxWithLoss):
+                t = Layer.SoftmaxWithLoss()
+            elif isinstance(t, Layer.IdentityWithLoss):
+                t = Layer.IdentityWithLoss()
+            elif isinstance(t, Layer.BatchNormalization):
+                t = Layer.BatchNormalization()
+
+
+            elif isinstance(t, Layer.Pooling):
+
+                t = Layer.Pooling(t.pool_h, t.pool_w, t.stride, t.pad)
+
+            s = jsp.encode(t)
             layer_info.append(s)
 
-        save_data["LayerInfo"] = layer_info
-
-        for layer_idx, layer in enumerate(self.layers):
-
+        for layer_name, layer in self.layers.items():
             if isinstance(layer, Layer.BatchNormalization):
-                params[f"BN_m{layer_idx}"] = layer.running_mean
-                params[f"BN_v{layer_idx}"] = layer.running_var
+                layer_idx = int(re.findall(r'[0-9]+', layer_name)[-1])
+                other_params[f"BN_m{layer_idx}"] = layer.running_mean
+                other_params[f"BN_v{layer_idx}"] = layer.running_var
+
+
+
+        save_data["LayerInfo"] = layer_info
+        # for layer_idx, layer in enumerate(self.layers.values()):
+        #     if isinstance(layer, Layer.BatchNormalization):
+        #         params[f"BN_m{layer_idx}"] = layer.running_mean
+        #         params[f"BN_v{layer_idx}"] = layer.running_var
 
         save_data["Params"] = params
+        save_data["OtherParams"] = other_params
 
         if path is None:
             path = f"train_weight_{str(datetime.datetime.now())[:-7].replace(':', '')}.npz"
+        # print(save_data)
 
         np.savez_compressed(path, **save_data)
         print(f"Weight was saved at {path}")
 
     def load_model(self, path):
-
+        import numpy as np2
         print(f"Loading Model... {path}")
-        load_data = np.load(path, allow_pickle=True)
+        load_data = np2.load(path, allow_pickle=True)
         other_config_data = dict(load_data.get("NetConfig").item())
         self.__init__(**other_config_data)
         layer_info = load_data.get("LayerInfo")
         for layer_encoded in layer_info:
             layer = jsp.decode(layer_encoded)
+            print(layer_encoded)
+
             self.add_layer(layer)
 
         self.params = load_data.get("Params").item()
+        other_params = load_data.get("OtherParams").item()
 
         for param_name, param_value in self.params.items():
+            print(param_name)
+
             layer_idx = int(re.findall(r'[0-9]+', param_name)[0])
             param_kind = re.sub(r'[0-9]+', '', param_name)
             target_layer = list(self.layers.items())[layer_idx][1]
@@ -473,6 +540,20 @@ class MultiLayerNet:
             elif param_kind == "BN_v":
                 target_layer.running_var = param_value
 
+        for param_name, param_value in other_params.items():
+            layer_idx = int(re.findall(r'[0-9]+', param_name)[0])
+            param_kind = re.sub(r'[0-9]+', '', param_name)
+            target_layer = list(self.layers.items())[layer_idx][1]
+            if param_kind == "BN_m":
+                target_layer.running_mean = param_value
+
+                print(target_layer)
+                print(target_layer.running_mean)
+            elif param_kind == "BN_v":
+                target_layer.running_var = param_value
+
+
+
 # %%
 
 
